#include <mpi.h>
#include <string.h>
#include "hpmtsolver.h"
#include "hplssolver.h"
#include "backupsol.h"
#include "mtglvec.h"
#include "global.h"
#include "globmat.h"
#include "hpglobal.h"
#include "seqfilesm.h"
#include "mechprint.h"
#include "matrix.h"
#include "vector.h"
#include "gmatrix.h"
#include "gtopology.h"
#include "dloadcase.h"
#include "elemswitch.h"
#include "nssolver.h"
#include "newtonraph.h"
#include "mathem.h"
// for evaluation of hypoplasticity ???!!!
#include "intpoints.h"
//#include "hypoplunsatexptherm.h"
#include "hypoplunsatexptherm2.h"
//#include "hypoplunsatexptherm2.h"
#include "slipsurf.h"


#ifndef FNAMELEN
 #define FNAMELEN 1001
#endif



/**
  The function calls solver of time dependent problems (viscoplasticity)
  for each load case.

  @return The function does not return anything.

  Created by TKr 27/07/2022 accroding to JK
*/
void parallel_homogenization_solve_time_dep_prob ()
{
  long lcid;
  
  //  load case must be equal to zero, no superposition can be used
  //  in this type of analysis
  lcid=0;
  
  //  solver of the problem
  parallel_homogenization_visco_solver2 (lcid);
}



/**
  Function solves time dependent problem (viscoplasticity).
  It uses one_step concept instead of standard one.
  
  @param lcid - load case id
   
  @return The function does not return anything.
  
  Created by Tomas Krejci 27/07/2022 according to Tomas Koudelka
*/
void parallel_homogenization_visco_solver2 (long lcid)
{
  long i, j,k,ii,jj,kk, li, nsts, ret, rest_calc,nelidom,buffsize1,buffsize2,ncomp,dim;
  double time, dt, dtmin, dtmax, dtr, end_time, prev_time;
  mt_glob_vec mt_gv;
  long *neldom,*buff;
  double *buff1,*buff2;
  MPI_Status stat;
  
  //  allocation of buffer
  //  buff[0] - the number of connections between the macroproblem and microproblem
  //  buff[1] - the maximum number of connections between the macroproblem and microproblem
  buff = new long [2];
  
  if (Myrank==0){  
    switch (Mp->mami){
    case elem_domain:{
      //  each element is connected with its own microproblem
      buff[0]=1;
      buff[1]=1;
      
      break;
    }
    case aggregate_domain:{
      //  generation of auxiliary data needed in parallel computation
      //  elements of macroproblem are aggregated, each finite element
      //  sends its data to microproblem and obtains its matrices
      
      //  map between finite elements of macroproblem and microproblems
      //  this map is read in the subroutine transtop::read_seq_top
      //  eldom[i]=j - the i-th finite element of macroproblem is connected to the j-th microproblem
      
      //  the number of elements on subdomains
      //  it is the number of macroelements connected to one microproblem
      //  the number of subdomains--microproblems must be Nproc-1
      //  neldom[i]=j - j finite elements of the macroproblem are connected to the i-th microproblem
      //  neldom[0]=0 - the master processor (myrank=0) contains no microproblem
      neldom = new long [Nproc];
      for (i=0;i<Nproc;i++){
	neldom[i]=0;
      }
      for (i=0;i<Mt->ne;i++){
	neldom[Gtm->stop->eldom[i]]++;
      }
      buffsize1=0;
      for (i=0;i<Nproc;i++){
	if (buffsize1<neldom[i])
	  buffsize1=neldom[i];
      }
      
      buff[1]=buffsize1;
      
      break;
    }
    case material_aggregate_domain:{
      //one domain represents one material = one processor
      //the aggregate covers elements of one materyal type
      neldom = new long [Nproc];
      for (i=0;i<Nproc;i++){
	neldom[i]=0;
      }
      
      for (i=0;i<Mt->ne;i++)
	neldom[Gtm->stop->eldom[i]] = 1;
      
      buff[0]=1;
      buff[1]=1;
      
      break;
    }
    case eff_aggregate_domain:{

      neldom = new long [Nproc];
      for (i=0;i<Nproc;i++){
	neldom[i]=0;
      }

      for (i=0;i<Mt->ne;i++){
	if(Gtm->stop->eldom[i] != Mt->elements[i].idm[0])
	  neldom[Gtt->stop->eldom[i]]++;
      }
      
      //  the aggregate sends averaged data to microproblem
      //  only aggregate sends and receives data, the same
      //  matrices are used on all elements in the aggregate
      buff[0]=1;
      buff[1]=1;
      
      break;
    }
    default:{
      par_print_err(Myrank,"unknown type of aggregation is required",__FILE__,__LINE__,__func__);
    }
    }    
  
    for (i=1;i<Nproc;i++){
      if (elem_domain == aggregate_domain)
	buff[0]=neldom[i];

      MPI_Send (buff,2,MPI_LONG,i,Myrank,MPI_COMM_WORLD);
      //MPI_Send_mine (buff,2,MPI_LONG,i,Myrank,MPI_COMM_WORLD);
    }
    if (elem_domain == aggregate_domain)
      buff[0]=neldom[0];
  }
  else{
    MPI_Recv (buff,2,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    //MPI_Recv_mine (buff,2,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  

  //  the number of connections between the macroproblem and microproblem
  //  the number of entities (elements or subdomains) of macroproblem connected with microproblem
  nelidom=buff[0];
  //  the maximum number of connections between the macroproblem and microproblem
  //  the maximum number of entities (elements or subdomains) of macroproblem connected with microproblem
  buffsize1=buff[1];
  
  //  dimension of the problem, one, two or three dimensional case
  //  the dimension is defined with respect to the first finite element
  dim = Mt->give_dimension (0);
  //  the number of stress/strain components
  ncomp = Mt->give_ncomp (0,0);
  
  buffsize2=buffsize1;
  //  buffer for displacments and strain components
  buffsize1*=(dim+ncomp+1);
  //  buffer for the stiffness matrix
  buffsize2*=(ncomp*ncomp);
  
  fprintf (stdout,"\n procesor %d  nelidom   %ld",Myrank,nelidom);
  fprintf (stdout,"\n procesor %d  buffsize1 %ld",Myrank,buffsize1);
  fprintf (stdout,"\n procesor %d  buffsize2 %ld",Myrank,buffsize2);
  
  
  delete [] buff;
  buff1 = new double [buffsize1];
  nullv (buff1,buffsize1);
  buff2 = new double [buffsize2];
  nullv (buff2,buffsize2);
  
  
  /**************************************************************************/

  if (Myrank==0){

    //
    //  initialization phase
    //
    //parallel_homogenization_visco_solver_init(lcid, rest_calc, mt_gv);
    visco_solver_init(lcid, rest_calc, mt_gv);
    
    // store initial value for attained time in the previous time step
    // (prev_time is exploited in the growing mechanical problems)
    prev_time = Mp->time;
    
    //  initial time increment
    dt = Mp->timecon.initialtimeincr ();
    //  end time
    end_time = Mp->timecon.endtime ();
    
    //  number of step
    li = i = mt_gv.istep;
    //  number of successful time steps (number of steps without inner iterations)
    nsts=0;
    // return status from the one_step function
    ret = 0;
    
    if (Mp->tprob == growing_mech_structure)  
      //saving values for the first iteration
      Mt->save_nodval(lcid);
  }
  /**************************************************************************/
  
  if (Myrank==0){
    for (ii=1;ii<Nproc;ii++){
      MPI_Send (&end_time,1,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
      //MPI_Send_mine (&end_time,1,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
    }
  }
  else{
    MPI_Recv (&end_time,1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    //MPI_Recv_mine (&end_time,1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  
  fprintf (stdout,"\n procesor %d end_time %lf",Myrank,end_time);


  // ***************************
  //   main iteration loop  ****
  // ***************************
  do
  {
    if (Myrank==0){
      //  new step number
      i++;
      
      // perform one time step
      time=Mp->timecon.newtime(dt);
      if (time<0.0)
	break;
      
      //  minimum time increment
      dtmin=Mp->timecon.dtmin;
      //  maximum time increment
      dtmax=Mp->timecon.dtmax;
      Mp->dtime = dt;
      if ((Mp->tprob == growing_mech_structure) && (prev_time == Mp->timecon.starttime()))
	prev_time = time;
      
      for (ii=1;ii<Nproc;ii++){
	MPI_Send (&time,1,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
	//MPI_Send_mine (&time,1,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Recv (&time,1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      //MPI_Recv_mine (&time,1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    MPI_Barrier (MPI_COMM_WORLD);

    
    /**************************************************************************/
    
    if (Myrank==0){
      
      // **********************************
      //  solution of microscale problems
      // **********************************
      for (ii=1;ii<Nproc;ii++){
	
	//  function assembles data from the macroproblem for the ii-th microproblem to the array buff1
	// tady pole buff1 musi mit rozmer podle nelidom??!! nebo podle poctu elementu??!!
	higher_to_lower_levelm (ii,buffsize1,buff1);
	
	MPI_Send (buff1,buffsize1,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
	//MPI_Send_mine (buff1,buffsize1,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Recv (buff1,buffsize1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      //MPI_Recv_mine(buff1,buffsize1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      for (ii=0;ii<nelidom;ii++){
	
	/* fprintf (Out,"\n\n *********************************************");
	   fprintf (Out,"\n Myrank  = %ld",Myrank);
	   fprintf (Out,"\n nelidom = %ld",ii);
	   fprintf (Out,"\n\n pred vypoctem paral_mechanical_homogenization:");
	   for (long ijk=0;ijk<buffsize1;ijk++){
	   //for (long ijk=0;ijk<1;ijk++){
	   fprintf (Out,"\n buff1 %ld   %le",ijk,buff1[ijk]);
	   }
	   
	   fprintf (Out,"\n");
	   for (long ijk=0;ijk<buffsize2;ijk++){
	   //for (long ijk=0;ijk<1;ijk++){
	   fprintf (Out,"\n buff2 %ld   %le",ijk,buff2[ijk]);
	   }
	   fflush(Out);
	*/

	paral_mechanical_homogenization (buff1+ii*(dim+ncomp+1),buff2+ii*(ncomp*ncomp));
	
	/* fprintf (Out,"\n\n po vypoctu paral_mechanical_homogenization:");
	   for (long ijk=0;ijk<buffsize1;ijk++){
	   //for (long ijk=0;ijk<1;ijk++){
	   fprintf (Out,"\n buff1 %ld   %le",ijk,buff1[ijk]);
	   }
	   fprintf (Out,"\n");
	   for (long ijk=0;ijk<buffsize2;ijk++){
	   //for (long ijk=0;ijk<1;ijk++){
	   fprintf (Out,"\n buff2 %ld   %le",ijk,buff2[ijk]);
	   }
	   fflush(Out);
	*/
      }
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    /**************************************************************************/
    
    if (Myrank==0){
      for (ii=1;ii<Nproc;ii++){
	MPI_Recv (buff2,buffsize2,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	//MPI_Recv_mine(buff2,buffsize2,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	
      	k=stat.MPI_TAG;
	jj=0;
	
	for (j=0;j<Mt->ne;j++){
	  
	  if (Gtm->stop->eldom[j]==k){
	    //  only elements connected with the k-th microscale are assumed
	  
	    /* fprintf (Out,"\n");
	       fprintf (Out,"\n Myrank  = %ld",Myrank);
	       fprintf (Out,"\n element no. = %ld",j);
	       fprintf (Out,"\n Gtm->stop->eldom[j] = %ld",Gtm->stop->eldom[j]);
	       fprintf (Out,"\n ii = %ld",ii);
	       fprintf (Out,"\n k  =stat.MPI_TAG = %ld",k);
	       fprintf (Out,"\n");
	       fflush(Out);
	    */

	    switch (Mp->mami){
	    case elem_domain:{
	      //  each element is connected with its own microproblem

	      Mm->hommatm[j].assemble_matrices (buff2,ncomp,dim);
	      break;
	    }
	      
	    case aggregate_domain:{
	      //  generation of auxiliary data needed in parallel computation
	      Mm->hommatm[j].assemble_matrices (buff2+jj*(ncomp*ncomp),ncomp,dim);
	      jj++;
	      break;
	    }
	    case material_aggregate_domain:{
	      //  aggregate sends averaged data to microproblem
	      //  index jj is not incremented in this case because all elements in aggregate??!!
	      //  obtain the same conductivity and capacity matrices
	      //
	      //  index of subdomain-aggregate = index of material (processor)
	      
	      //fprintf (Outt,"\n Myrank  = %ld",Myrank);
	      //fprintf (Outt,"\n from Myrank  = %ld",ii);
	      //fprintf (Outt,"\n element  = %ld",j);
	      //fprintf (Outt,"\n\n pred ulozenim matic na materialu c. %ld:",j);
	      //for (long ijk=0;ijk<buffsize2;ijk++){
	      //fprintf (Outt,"\n buff2 %ld   %le",ijk,buff2[ijk]);
	      //}
	      
	      Mm->hommatm[j].assemble_matrices (buff2,ncomp,dim);

	      //fflush(Outt);
	      break;
	    }
	    case eff_aggregate_domain:{
	      //  aggregate sends averaged data to microproblem
	      //  index jj is not incremented in this case because all elements in aggregate
	      //  obtain the same conductivity and capacity matrices
	      //
	      //  index of subdomain-aggregate = index of element material
	      kk = Mt->elements[i].idm[0];
	      Mm->hommatm[kk].assemble_matrices (buff2+kk*(ncomp*ncomp),ncomp,dim);
	      break;
	    }
	      
	    default:{
	      par_print_err(Myrank,"unknown type of aggregation is required",__FILE__,__LINE__,__func__);
	    }
	    } 
	  }
	}//  end loop over the number of elements
      }//  end loop over the number of processors
    }//  end of the master processor part
    else{
      MPI_Send (buff2,buffsize2,MPI_DOUBLE,0,Myrank,MPI_COMM_WORLD);
      //MPI_Send_mine (buff2,buffsize2,MPI_DOUBLE,0,Myrank,MPI_COMM_WORLD);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    fflush (stdout);
   

    /**************************************************************************/
    
    if (Myrank==0){
           
      // *********************************
      //  solution of macroscale problem
      // *********************************
      
      //ret = parallel_homogenization_one_step(lcid, time, dt, dtr, prev_time, rest_calc, i, li, mt_gv)
      ret = one_step(lcid, time, dt, dtr, prev_time, rest_calc, i, li, mt_gv);
      
      if (ret >= 0) // equilibrium was attained
	{
	  
	  //  Backup of r is used in case that the Newton-Raphson procedure does not converge.
	  //  lhsb[j]=r[j]
	  copyv(mt_gv.r, mt_gv.lhsb, Ndofm);
	  
	  // actual total load vector becomes previous total load vector
	  // fp[k]=f[k]
	  copyv(mt_gv.f, mt_gv.fp, Ndofm);
	  
	  
	  if (Mp->tprob == growing_mech_structure)
	    {
	      // Following two commands must be performed here because in case that
	      // elements or nodes are changed in the next step and the step would not converge,
	      // the nodal values and nodal forces could not be recovered correctly
	      // nodval due to changes in Mt->nodedispl and nodforce due to changes in vector f
	      
	      // backup of nodal values because code numbers may be changed due to
	      Mt->save_nodval(lcid);
	      //  backup of nodal forces because code numbers may be changed
	      Mt->save_nodforce(mt_gv.f);
	      
	      // new elements can be definitely changed to old ones
	      Gtm->update_auxinf();
	      prev_time = Mp->time;
	    }
	  
	  //  update of values stored at integration points
	  Mm->updateipval();
	  compute_req_val (lcid);
	  print_step(lcid, Mp->istep, Mp->time, mt_gv.fl);
	  print_flush();
	  
	  if ((Mp->timecon.isitimptime()==1) && (Mp->hdbcont.save_stat()))
	    {
	      if (Mespr==1)
		fprintf (stdout,"\n Creating MEFEL backup file\n");
	      solver_save (mt_gv.r,mt_gv.fp,Mp->istep,Mp->time,dt,&Mp->timecon,Ndofm);
	    }
	  if (ret == 0)
	    nsts++;      
	  if (nsts==2)
	    {
	      if (dt < dtmax)
		{
		  dt*=2.0;
		  if (dt>dtmax)//maximum time increment
		    dt = dtmax;
		  nsts=0;
		  if (Mespr==1)  
		    fprintf (stdout,"\n\n time increment is enlarged because no inner loop was neccessary in previous 2 steps");
		}
	    }
	}
      else
	{
	  //  reduction of the time increment because
	  //  inner loop was not able to enforce equilibrium
	  nsts = 0;
	  
	  if (dt > dtmin){
	    // time step can be reduced
	    if (dtr < 1.0)
	      dt *= dtr;  // use reduction factor proposed by material models if they proposed ever reduction
	    else
	      dt/=2.0;   // use general reduction factor
	    
	    // check the minumum length of time step
	    if (dt < dtmin)
	      dt = dtmin;
	    
	    Mp->timecon.oldtime ();
	    i--;
	    
	    // copy state variables from the last attained equilibrium state (eqother array) to the actual state variables (other array)
	    Mm->updateother();
	    
	    if (Mespr==1) {
	      if (dt == dtmin)
		fprintf (stdout,"\n\n time increment is reduced because the inner loop was not able to enforce equilibrium\n");
	      else
		fprintf (stdout,"\n\n time increment is reduced because the inner loop was not able to enforce equilibrium");
	    }
	    // change status of elements and nodes according to the previous time step
	    if (Mp->tprob == growing_mech_structure){
	      update_elemnod_prev_time(prev_time, mt_gv.ncd, mt_gv.nce);
	    }
	  }
	  else{
	    // time step should be reduced but it was attained the minimum time step length
	    if (Mespr==1)  fprintf (stderr,"\n\n reduced time increment is less than minimum time increment");
	    if (Mespr==1)  fprintf (stderr,"\n computation failed (file %s, line %d)\n",__FILE__,__LINE__);
	    if (Mespr==1)  fprintf (stderr,"\n FORCED output of results from this step is performed\n");
	    //  update of values stored at integration points
	    Mm->updateipval();
	    compute_req_val (lcid);
	    print_step_forced(lcid, i, Mp->time, mt_gv.fb);
	    print_flush();
	    break;
	  }
	}
    }

    fprintf (stdout,"\n procesor %d time = %lf end_time = %lf",Myrank,time,end_time);
    fflush (stdout);
    
    /**************************************************************************/
    
  }while(Mp->time<end_time);

  if (Myrank==0){
    print_close ();
  }

  delete [] buff1;
  delete [] buff2;
}



/**
  Function allocates and initializes global vectors used in mtsolver
  It is used in one_step concept solver.

  @param lcid      - load case id
  @param rest_calc - indicator of calculation restorage from backup 
                     (0=no restorage was required, 1=restorage was performed)
  @param mt_gv     - structure with global vectors used in the solver

  Created by Tomas Krejci 27/07/2022 according to Tomas Koudelka
*/
void parallel_homogenization_visco_solver_init(long lcid, long &rest_calc, mt_glob_vec &mt_gv)
{
  long n;
  double dt;  
  //  number of mechanical degrees of freedom
  n=Ndofm;

  mt_gv.alloc(n);
  //  vector of nodal displacements
  mt_gv.r = Lsrs->give_lhs (lcid);
  //  vector of nodal prescribed forces
  mt_gv.f = Lsrs->give_rhs (lcid);

  //  initial time
  Mp->time=Mp->timecon.starttime ();
  Mp->nlman->divc_step = new double[10];
  Mp->nlman->divc_err  = new double[10];

  rest_calc = 0;

  //  inital printing of output and graphical informations
  if (Mp->hdbcont.restore_stat())
  {
    // initiation of mechanical material models
    Mm->initmaterialmodels(lcid);
    // initiation of material models on auxiliary integration points
    Mm->aip_initmaterialmodels(lcid);
    if (Mespr==1)
      fprintf (stdout,"\n Reading of MEFEL backup file\n");
    solver_restore (mt_gv.r, mt_gv.fp, mt_gv.istep, Mp->time, dt, &Mp->timecon, n);
    Mm->updateother();

    //  Backup of r is used in case that the Newton-Raphson procedure does not converge.
    //  lhsb[j]=r[j]
    copyv(mt_gv.r, mt_gv.lhsb, n);

    if (Mp->tprob == growing_mech_structure)    
      update_elnod_stat_after_hdbrest(lcid);

    // update actual material index for all elements
    Mm->update_actual_mat_id(lcid, 0);
    // set indicator of calculation started from restored backup data
    rest_calc = 1;
    // reopen output files
    print_init(-1, "at");
  }
  else
  {
    mt_gv.istep=0;
    //  initiation of mechanical material models
    Mm->initmaterialmodels(lcid);
    // initiation of material models on auxiliary integration points
    Mm->aip_initmaterialmodels(lcid);
    // compute stresses from eigenstrains or initial irreversible strains
    if ((Mp->eigstrains > 0) && (Mp->eigstrains < 4))
      internal_forces(lcid, mt_gv.fp);
    // print initial step to the output files
    print_init(-1, "wt");
    print_step(lcid, mt_gv.istep, Mp->time, mt_gv.f);
    print_flush();
  }
}


/**
  Function executes one step of time dependent mechanical algorithm
  (visco-elasticity, visco-plasticity, etc.)
   
  @param lcid - load case id
  @param time - actual time
  @param dt   - actual time increment
  @param dtr   - the minimum ratio of required time step size to the actual one required by material models (output)
  @param prev_time - attained time in the previous time step (needed only for growing structures)
  @param rest_calc - inidicator whether step follows restorage from the backup immediately (=1) or no (=0)
  @param istep - time step id
  @param li    - time step id of the first performed time step
  @param mt_gv - structure with pointers to %vectors of righthand and lefthand side

  @retval  0 - if the equilibrium was attained with NO inner loop
  @retval  >0 - if the equilibrium was attained and inner loop WAS performed
  @retval -1 - if the equilibrium was NOT attained

  Created by Tomas Krejci 27/07/2022 according to Tomas Koudelka and Jaroslav Kruis
*/
long parallel_homogenization_one_step (long lcid,double time, double dt, double &dtr, double prev_time, long rest_calc, long istep, long li, mt_glob_vec &mt_gv)
{
  long j,n,ini,um, mncd, mnce, mnae;
  double *f,*fl,*fi,*fb,*fp,*r,*dr,*lhsb, *flp;
  long *ifn;
  double norf, err;
  matrix lsm_a(3,3);
  vector lsm_r(3), lsm_l(3);
  answertype fusm;

  //new time increment
  Mp->time = time;
  //  new step number
  Mp->istep = istep;
  Mp->jstep = -1;
  //  indicator of strain computation
  //  it means, strains at integration points have not been computed
  Mp->strainstate=0;
  //  indicator of stress computation
  //  it means, stresses at integration points have not been computed
  Mp->stressstate=0;
  //  indicator of computation of other array
  //  it means, stresses at integration points have not been computed
  Mp->otherstate=0;

  //  vector of nodal displacements
  r  = mt_gv.r;
  //  vector of nodal prescribed forces
  f  = mt_gv.f;
  //  vector of increments of nodal displacements
  dr = mt_gv.dr;
  //  vector of prescribed forces from the previous step
  fp = mt_gv.fp;
  //  vector of prescribed forces due to force load from the previous step
  flp = mt_gv.flp;
  //  vector of prescribed force loads, it does not contain forces caused by temperature, etc.
  fl = mt_gv.fl;
  //  vector of internal forces
  fi = mt_gv.fi;
  //  auxiliary force vector
  fb = mt_gv.fb;
  //  backup of the nodal displacements
  lhsb = mt_gv.lhsb;
  // indicators of nodes on interfaces between old and new parts of the structure
  ifn = mt_gv.ifn;

  //  number of mechanical degrees of freedom
  n=Ndofm;

  // update actual material index for all old elements and call initval for those whose indeces have changed
  if (rest_calc == 0)
  {
    um = Mm->update_actual_mat_id(lcid, 1);
    if (um)   fusm = yes;
    else      fusm = no;
  }  
  else  // do not update if the calculation was restored from HD
    fusm = no;

  // update statuses of all nodes, DOFS and elements for growing structures,
  // calculation of initial displacements of new parts of the structure
  if (Mp->tprob == growing_mech_structure)
  {    
    update_elnod_stat(lcid, istep, prev_time, ifn, r, fb, fp, mnce, mncd, mnae);
    mt_gv.ncd = mncd;
    mt_gv.nce = mnce;
    n = Ndofm;
  }

  //  maximum number of iterations in inner loop
  ini = Mp->nlman->niilnr;
  //  required norm of vector of unbalanced forces
  err = Mp->nlman->errnr;

  if (Mespr==1)  fprintf (stdout,"\n\n ------------------------------------------------------------------------");
  if (Mespr==1)  fprintf (stdout,"\n MEFEL Time step = %ld,  Time %e,  Time increment = %e",istep,Mp->time,dt);
  if (Mespr==1)  fprintf (stdout,"\n ------------------------------------------------------------------------\n");

  //  assembling of vectors of prescribed nodal forces
  mefel_right_hand_side (lcid,f,fl);
  
  //  computation of sums of force components in particular directions
  //  it is used in special creep and consolidation models
  Mb->comp_sum (f);

  // assemble force vector due to removed elements and apply it gradually if required
  if (Mp->tprob == growing_mech_structure)
  {    
    forces_due2removed_elem(lcid, mnce-mnae, istep, prev_time, fi, fp, flp, fb, r);
  }

  //  increments of prescribed internal forces
  incr_internal_forces (lcid,fi);

  // assembling of right hand side
  // fb[j]=f[j]-fp[j]+fi[j]
  subv(f, fp, fb, n);
  addv(fb, fi, fb, n);

  // perform one load step with the help of Newton-Raphson procedure
  j = 0;  

  norf = gnewton_raphson_one_step(lcid, Mp->nlman, fl, r, fb, dr, fi, dtr, istep-1, j, li, fusm);

  if (norf>err)
  {
    // equilibrium state was not attained
    if (Mespr==1) fprintf (stdout,"\n\n Iteration process does not converge req. err = %e, reached err = %e", err,norf);
    
    //  backup is used for actual values
    //  r[k]=lhsb[k]
    copyv(lhsb, r, n);
    
    return -1;
  }

  // equilibrium state was attained
  if ((Mespr==1) && (j > 0))  fprintf (stdout,"\n\n Last inner iteration step number j = %ld,     norf = %e \n\n",j,norf);
  
  // print resulting vector of internal forces
  Mb->comp_sum (fi);

  if (j < 0)
    return 0;
  
  if (Mp->jstep > 1)
    return 1;
  else
    return 0;
}
