#include "mpi.h"
#include <math.h>
#include <string.h>
#include "pcsolver.h"
#include "globalc.h"
#include "global.h"
#include "globalt.h"
#include "globmat.h"
#include "globmatt.h"
#include "globmatc.h"
#include "elemswitcht.h"
#include "gmatrix.h"
#include "gtopology.h"
#include "mechprint.h"
#include "transprint.h"
#include "backupsol.h"
#include "backupsolt.h"
#include "mtsolver.h"
#include "npsolvert.h"
#include "dnnpsolvert.h"
#include "hpmtsolver.h"
#include "hppcsolver.h"
#include "hplssolver.h"
#include "hpnpsolvert.h"
#include "hpglobalc.h"

/**
   function solves partially coupled thermo-hydro-mechanical time-dependent problem   
   transport problems are linear or nonlinear
   mechanical analysis is always nonlinear
*/
void parallel_homogenization_solve_pcouplprob ()
{

  switch (Cp->tnlinsol){
  case newtonc:{
    parallel_homogenization_newton_raphson_parcoupl_common_dt (0);
    break;
  }
  default:{
    print_err("unknown solver of nonlinear equation system is required",__FILE__,__LINE__,__func__);
  }
  }
}



/**
   function solves partially coupled thermo-hydro-mechanical time-dependent problem
   by the Newton-Raphson method
   
   @param lcid - load case id
   
   TKr 29/07/2022 according to TKo
*/
void parallel_homogenization_newton_raphson_parcoupl_common_dt (long lcid)
{
  long mefel_convergence,trfel_convergence,stop,nelidomm,nelidomt,buffsize1,buffsize2,buffsize3,buffsize4,ntm,ncomp,dimt,dimm;
  double dt,time,end_time;
  long i,ii,jj,j,k,kk,ti,tli,tnsts, btnsts, tret;
  double tnewtime, tdt;
  np_glob_vec np_gv;
  long tlcid = 0;//zatim??!!
  long mi, mli, mnsts, bmnsts, mret;
  double mnewtime, mdt, mdtr;
  mt_glob_vec mt_gv;
  long mlcid = 0;//zatim??!!
  long rest_calcm, rest_calct;
  double prev_timem, prev_timet;
  long *neldomm,*neldomt,*buff;
  double *buff1,*buff2,*buff3,*buff4;
  MPI_Status stat;
  
  //  allocation of buffer
  //  buff[0] - the number of connections between the macroproblem and microproblem for transport and mechanical part
  //  buff[1] - the maximum number of connections between the macroproblem and microproblem for transport and mechanical part
  buff = new long [2];

  // *************************
  // ***** TRANSPORT PART ****
  if (Myrank==0){  
    switch (Tp->mami){
    case elem_domain:{
      //  each element is connected with its own microproblem
      buff[0]=1;
      buff[1]=1;
            
      break;
    }
    case aggregate_domain:{
      //  generation of auxiliary data needed in parallel computation
      //  elements of macroproblem are aggregated, each finite element
      //  sends its data to microproblem and obtains its matrices
      
      //  map between finite elements of macroproblem and microproblems
      //  this map is read in the subroutine transtop::read_seq_top
      //  eldom[i]=j - the i-th finite element of macroproblem is connected to the j-th microproblem
      
      //  the number of elements on subdomains
      //  it is the number of macroelements connected to one microproblem
      //  the number of subdomains--microproblems must be Nproc-1
      //  neldomt[i]=j - j finite elements of the macroproblem are connected to the i-th microproblem
      //  neldomt[0]=0 - the master processor (myrank=0) contains no microproblem
      neldomt = new long [Nproc];
      for (i=0;i<Nproc;i++){
	neldomt[i]=0;
      }
      for (i=0;i<Tt->ne;i++){
	neldomt[Gtt->stop->eldom[i]]++;
      }
      buffsize1=0;
      for (i=0;i<Nproc;i++){
	if (buffsize1<neldomt[i])
	  buffsize1=neldomt[i];
      }
      
      buff[1]=buffsize1;
      
      break;
    }
    case material_aggregate_domain:{
      //one domain represents one material = one processor
      //the aggregate covers elements of one materyal type
      neldomt = new long [Nproc];
      for (i=0;i<Nproc;i++){
	neldomt[i]=0;
      }
      
      for (i=0;i<Tt->ne;i++)
	neldomt[Gtt->stop->eldom[i]] = 1;
      
      buff[0]=1;
      buff[1]=1;
      
      break;
    }
    case eff_aggregate_domain:{

      neldomt = new long [Nproc];
      for (i=0;i<Nproc;i++){
	neldomt[i]=0;
      }

      for (i=0;i<Tt->ne;i++){
	if(Gtt->stop->eldom[i] != Tt->elements[i].idm[0])
	  neldomt[Gtt->stop->eldom[i]]++;
      }
      
      //  the aggregate sends averaged data to microproblem
      //  only aggregate sends and receives data, the same
      //  matrices are used on all elements in the aggregate
      buff[0]=1;
      buff[1]=1;
      
      break;
    }
    default:{
      par_print_err(Myrank,"unknown type of aggregation is required",__FILE__,__LINE__,__func__);
    }
    }    
  
    for (i=1;i<Nproc;i++){
      if (elem_domain == aggregate_domain)
	buff[0]=neldomt[i];

      MPI_Send (buff,2,MPI_LONG,i,Myrank,MPI_COMM_WORLD);
      //MPI_Send_mine (buff,2,MPI_LONG,i,Myrank,MPI_COMM_WORLD);
    }
    if (elem_domain == aggregate_domain)
      buff[0]=neldomt[0];
  }
  else{
    MPI_Recv (buff,2,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    //MPI_Recv_mine (buff,2,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  

  //  the number of connections between the macroproblem and microproblem
  //  the number of entities (elements or subdomains) of macroproblem connected with microproblem
  nelidomt=buff[0];
  //  the maximum number of connections between the macroproblem and microproblem
  //  the maximum number of entities (elements or subdomains) of macroproblem connected with microproblem
  buffsize1=buff[1];
  
  //  dimension of the problem, one, two or three dimensional case
  //  the dimension is defined with respect to the first finite element
  dimt = Tt->give_dimension (0);
  //  the number of transported materials
  ntm = Tp->ntm;
  
  buffsize2=buffsize1;
  //  buffer for values and gradient components
  buffsize1*=ntm*(dimt+1);
  //  buffer for the conductivity and capacity matrices
  //  diagonal capacity matrices are assumed
  buffsize2*=((ntm*dimt)*(ntm*dimt)+ntm*ntm);
  
  fprintf (stdout,"\n procesor %d  nelidomt   %ld",Myrank,nelidomt);
  fprintf (stdout,"\n procesor %d  buffsize1 %ld",Myrank,buffsize1);
  fprintf (stdout,"\n procesor %d  buffsize2 %ld",Myrank,buffsize2);
  
  
  buff1 = new double [buffsize1];
  nullv (buff1,buffsize1);
  buff2 = new double [buffsize2];
  nullv (buff2,buffsize2);

  
  // **************************
  // ***** MECHANICAL PART ****
  if (Myrank==0){  
    switch (Mp->mami){
    case elem_domain:{
      //  each element is connected with its own microproblem
      buff[0]=1;
      buff[1]=1;
      
      break;
    }
    case aggregate_domain:{
      //  generation of auxiliary data needed in parallel computation
      //  elements of macroproblem are aggregated, each finite element
      //  sends its data to microproblem and obtains its matrices
      
      //  map between finite elements of macroproblem and microproblems
      //  this map is read in the subroutine transtop::read_seq_top
      //  eldom[i]=j - the i-th finite element of macroproblem is connected to the j-th microproblem
      
      //  the number of elements on subdomains
      //  it is the number of macroelements connected to one microproblem
      //  the number of subdomains--microproblems must be Nproc-1
      //  neldomm[i]=j - j finite elements of the macroproblem are connected to the i-th microproblem
      //  neldomm[0]=0 - the master processor (myrank=0) contains no microproblem
      neldomm = new long [Nproc];
      for (i=0;i<Nproc;i++){
	neldomm[i]=0;
      }
      for (i=0;i<Mt->ne;i++){
	neldomm[Gtm->stop->eldom[i]]++;
      }
      buffsize3=0;
      for (i=0;i<Nproc;i++){
	if (buffsize3<neldomm[i])
	  buffsize3=neldomm[i];
      }
      
      buff[1]=buffsize3;
      
      break;
    }
    case material_aggregate_domain:{
      //one domain represents one material = one processor
      //the aggregate covers elements of one materyal type
      neldomm = new long [Nproc];
      for (i=0;i<Nproc;i++){
	neldomm[i]=0;
      }
      
      for (i=0;i<Mt->ne;i++)
	neldomm[Gtm->stop->eldom[i]] = 1;
      
      buff[0]=1;
      buff[1]=1;
      
      break;
    }
    case eff_aggregate_domain:{

      neldomm = new long [Nproc];
      for (i=0;i<Nproc;i++){
	neldomm[i]=0;
      }

      for (i=0;i<Mt->ne;i++){
	if(Gtm->stop->eldom[i] != Mt->elements[i].idm[0])
	  neldomm[Gtm->stop->eldom[i]]++;
      }
      
      //  the aggregate sends averaged data to microproblem
      //  only aggregate sends and receives data, the same
      //  matrices are used on all elements in the aggregate
      buff[0]=1;
      buff[1]=1;
      
      break;
    }
    default:{
      par_print_err(Myrank,"unknown type of aggregation is required",__FILE__,__LINE__,__func__);
    }
    }    
  
    for (i=1;i<Nproc;i++){
      if (elem_domain == aggregate_domain)
	buff[0]=neldomm[i];

      MPI_Send (buff,2,MPI_LONG,i,Myrank,MPI_COMM_WORLD);
      //MPI_Send_mine (buff,2,MPI_LONG,i,Myrank,MPI_COMM_WORLD);
    }
    if (elem_domain == aggregate_domain)
      buff[0]=neldomm[0];
  }
  else{
    MPI_Recv (buff,2,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    //MPI_Recv_mine (buff,2,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  

  //  the number of connections between the macroproblem and microproblem
  //  the number of entities (elements or subdomains) of macroproblem connected with microproblem
  nelidomm=buff[0];
  //  the maximum number of connections between the macroproblem and microproblem
  //  the maximum number of entities (elements or subdomains) of macroproblem connected with microproblem
  buffsize3=buff[1];
  
  //  dimension of the problem, one, two or three dimensional case
  //  the dimension is defined with respect to the first finite element
  dimm = Mt->give_dimension (0);
  //  the number of stress/strain components
  ncomp = Mt->give_ncomp (0,0);
  
  buffsize4=buffsize3;
  //  buffer for displacments and strain components
  buffsize3*=(dimm+ncomp+1);
  //  buffer for the stiffness matrix
  buffsize4*=(ncomp*ncomp);
  
  fprintf (stdout,"\n procesor %d  nelidomm   %ld",Myrank,nelidomm);
  fprintf (stdout,"\n procesor %d  buffsize3 %ld",Myrank,buffsize3);
  fprintf (stdout,"\n procesor %d  buffsize4 %ld",Myrank,buffsize4);
  
  
  delete [] buff;
  buff3 = new double [buffsize3];
  nullv (buff3,buffsize3);
  buff4 = new double [buffsize4];
  nullv (buff4,buffsize4);
  // **************************
  
  // **************************************************************************
  if (Myrank==0){
    //
    //  METR initialization phase
    //
    //  starting time
    Cp->time = Cp->timecon.starttime ();
    //  time increment
    dt = Cp->timecon.initialtimeincr ();
    //  end time
    end_time = Cp->timecon.endtime ();  
    
    Mp->time = Cp->time;
    Tp->time = Cp->time;
    //this part below is not neccessary because of independent time controlers of TRFEL and MEFEL
    //  mechanical time controller is rewritten by coupled time controller
    //Mp->timecon.take_values (Cp->timecon);
    //  transport time controller is rewritten by coupled time controller
    //Tp->timecont.take_values (Cp->timecon);
    // end this part
    
    
    //  initial approximation of quantities from TRFEL nodes to MEFEL integration points
    initapproximation();
    init_trfel_mefel ();
    approximation();
    
    //
    //  TRFEL initialization phase
    //
    if(Tp->tprob == discont_nonlin_nonstat_problem)
      nonstat_solver_dform_init(tlcid, np_gv);
    else
      nonstat_solver_init(tlcid, rest_calct, np_gv);
    //  initial time increment
    tdt = Tp->timecont.initialtimeincr ();
    //  number of step
    tli = ti = np_gv.istep;
    //  number of successful time steps (number of steps without inner iterations)
    tnsts = btnsts = 0;
    // return status from the one_step function
    tret = 0;
    // store initial value for attained time in the previous time step
    // (prev_time is exploited in the growing transport problems)
    prev_timet = Tp->time;
    
    if (Cp->dpt == pass_by_aux_ip){
      aip_approximation(Mm->tnip, MTipmap);
      actualize_aip_nonmechq(Tm->tnip, TMipmap);
    }
    
    //
    //  MEFEL initialization phase
    //
    //parallel_homogenization_visco_solver_init(mlcid, rest_calcm, mt_gv);
    visco_solver_init(mlcid, rest_calcm, mt_gv);
    
    //  initial time increment
    mdt = Mp->timecon.initialtimeincr ();
    //  number of step
    mli = mi = mt_gv.istep;
    //  number of successful time steps (number of steps without inner iterations)
    mnsts = bmnsts = 0;
    // return status from the one_step function
    mret = 0;
    // store initial value for attained time in the previous time step
    // (prev_time is exploited in the growing mechanical problems)
    prev_timem = Mp->time;
      
    // number of METR step
    i=0;  
    //setup of flag for successfully performed MEFEL step to yes
    mefel_convergence = 1;
    //setup of flag for successfully performed TRFEL step to yes
    trfel_convergence = 1;
    //  auxiliary indicator of convergence
    stop=0;
    
    //saving values for the first iteration
    if ((Tp->tprob == growing_np_problem) || (Tp->tprob == growing_np_problem_nonlin))
      Tt->lhs_save (np_gv.lhs,Lsrst->lhsi,np_gv.tdlhs);
    if (Mp->tprob == growing_mech_structure)  
      Mt->save_nodval(lcid);
    
    //  METR is copying data between TRFEL and MEFEL in both directions
    pass_coup_data(lcid);
  }

  // **************************************************************************

  if (Myrank==0){
    for (ii=1;ii<Nproc;ii++){
      MPI_Send (&end_time,1,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
      //MPI_Send_mine (&end_time,1,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
    }
  }
  else{
    MPI_Recv (&end_time,1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    //MPI_Recv_mine (&end_time,1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  
  fprintf (stdout,"\n procesor %d end_time %lf\n",Myrank,end_time);

  // ***************************
  //  main iteration loop  ****
  // ***************************
  
  do{
  
    if (Myrank==0){
      //  new step number
      i++;
      
      //  TRFEL time
      tnewtime = Tp->time = Tp->timecont.newtime (tdt);
      //  MEFEL time
      mnewtime = Mp->time = Mp->timecon.newtime (mdt);
      
      
      if (Tp->time > Mp->time)
	Cp->time = Mp->time;
      else
      Cp->time = Tp->time;

      time=Cp->time;
      
      //  time synchronization
      if (Tp->time > Mp->time){
	Tp->timecont.oldtime ();
	tnewtime = Tp->time = Tp->timecont.newtime (mdt);
	tdt = mdt; // actual TRFEL dt for linear solvers (due to recovery of decreased time step in MEFEL)
	if (btnsts)
	  tnsts = btnsts;
      }
      
      // (prev_time is exploited in the growing transport problems)
      if (((Tp->tprob == growing_np_problem) || (Tp->tprob == growing_np_problem_nonlin)) &&
	  (prev_timet == Cp->timecon.starttime()))
	prev_timet = Tp->time;
      
      //  TRFEL solution
      trfel_convergence = 0;
      
      ti++;
      
      for (ii=1;ii<Nproc;ii++){
	MPI_Send (&time,1,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
	//MPI_Send_mine (&time,1,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Recv (&time,1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      //MPI_Recv_mine (&time,1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    // **************************************************************************
     
    if (Myrank==0){
      // ******************************************
      //  solution of microscale TRANSPORT problems 
      // ******************************************
       for (ii=1;ii<Nproc;ii++){
	
	//  function assembles data from the macroproblem for the ii-th microproblem to the array buff1
	// tady pole buff1 musi mit rozmer podle nelidomt??!! nebo podle poctu elementu??!!
	higher_to_lower_level (ii,buffsize1,buff1);
	
	MPI_Send (buff1,buffsize1,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
	//MPI_Send_mine (buff1,buffsize1,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Recv (buff1,buffsize1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      //MPI_Recv_mine(buff1,buffsize1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      for (ii=0;ii<nelidomt;ii++){
	
	//fprintf (Outt,"\n\n *********************************************");
	//fprintf (Outt,"\n Myrank  = %ld",Myrank);
	//fprintf (Outt,"\n nelidomt = %ld",ii);
	//fprintf (Outt,"\n\n pred vypoctem paral_transport_homogenization:");
	//for (long ijk=0;ijk<buffsize1;ijk++){
	////for (long ijk=0;ijk<1;ijk++){
	//fprintf (Outt,"\n buff1 %ld   %le",ijk,buff1[ijk]);
	//}
	//fprintf (Outt,"\n");
	//for (long ijk=0;ijk<buffsize2;ijk++){
	////for (long ijk=0;ijk<1;ijk++){
	//fprintf (Outt,"\n buff2 %ld   %le",ijk,buff2[ijk]);
	//}
	//fflush(Outt);
	
	paral_transport_homogenization (buff1+ii*ntm*(dimt+1),buff2+ii*((ntm*dimt)*(ntm*dimt)+ntm*ntm),
					buff2+ii*((ntm*dimt)*(ntm*dimt)+ntm*ntm)+(ntm*dimt)*(ntm*dimt));
	
	//fprintf (Outt,"\n\n po vypoctu paral_transport_homogenization:");
	//for (long ijk=0;ijk<buffsize1;ijk++){
	////for (long ijk=0;ijk<1;ijk++){
	//fprintf (Outt,"\n buff1 %ld   %le",ijk,buff1[ijk]);
	//}
	//fprintf (Outt,"\n");
	//for (long ijk=0;ijk<buffsize2;ijk++){
	////for (long ijk=0;ijk<1;ijk++){
	//fprintf (Outt,"\n buff2 %ld   %le",ijk,buff2[ijk]);
	//}
	//fflush(Outt);
      }
    }
    MPI_Barrier (MPI_COMM_WORLD);
    // **************************************************************************

    if (Myrank==0){
      for (ii=1;ii<Nproc;ii++){
	MPI_Recv (buff2,buffsize2,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	//MPI_Recv_mine(buff2,buffsize2,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	
      	k=stat.MPI_TAG;
	jj=0;
	
	for (j=0;j<Tt->ne;j++){
	  
	  //fprintf (Outt,"\n Myrank  = %ld",Myrank);
	  //fprintf (Outt,"\n element no. = %ld",j);
	  //fprintf (Outt,"\n Gtt->stop->eldom[j] = %ld",Gtt->stop->eldom[j]);
	  //fprintf (Outt,"\n ii = %ld",ii);
	  //fprintf (Outt,"\n k  =stat.MPI_TAG = %ld",k);
	  
	  if (Gtt->stop->eldom[j]==k){
	    //  only elements connected with the k-th microscale are assumed

	    switch (Tp->mami){
	    case elem_domain:{
	      //  each element is connected with its own microproblem
	      Tm->hommat[j].assemble_matrices (buff2,ntm,dimt);
	      break;
	    }
	      
	    case aggregate_domain:{
	      //  generation of auxiliary data needed in parallel computation
	      Tm->hommat[j].assemble_matrices (buff2+jj*((ntm*dimt)*(ntm*dimt)+ntm*ntm),ntm,dimt);
	      jj++;
	      break;
	    }
	    case material_aggregate_domain:{
	      //  aggregate sends averaged data to microproblem
	      //  index jj is not incremented in this case because all elements in aggregate??!!
	      //  obtain the same conductivity and capacity matrices
	      //
	      //  index of subdomain-aggregate = index of material (processor)

	      //fprintf (Outt,"\n Myrank  = %ld",Myrank);
	      //fprintf (Outt,"\n from Myrank  = %ld",ii);
	      //fprintf (Outt,"\n element  = %ld",j);
	      //fprintf (Outt,"\n\n pred ulozenim matic na materialu c. %ld:",j);
	      //for (long ijk=0;ijk<buffsize2;ijk++){
	      //fprintf (Outt,"\n buff2 %ld   %le",ijk,buff2[ijk]);
	      //}
	      
	      Tm->hommat[j].assemble_matrices (buff2,ntm,dimt);

	      //fflush(Outt);
	      break;
	    }
	    case eff_aggregate_domain:{
	      //  aggregate sends averaged data to microproblem
	      //  index jj is not incremented in this case because all elements in aggregate
	      //  obtain the same conductivity and capacity matrices
	      //
	      //  index of subdomain-aggregate = index of element material
	      kk = Tt->elements[i].idm[0];
	      Tm->hommat[kk].assemble_matrices (buff2+kk*((ntm*dimt)*(ntm*dimt)+ntm*ntm),ntm,dimt);
	      break;
	    }
	      
	    default:{
	      par_print_err(Myrank,"unknown type of aggregation is required",__FILE__,__LINE__,__func__);
	    }
	    }
	    
	  }
	}//  end loop over the number of elements
      }//  end loop over the number of processors
    }//  end of the master processor part
    else{
      MPI_Send (buff2,buffsize2,MPI_DOUBLE,0,Myrank,MPI_COMM_WORLD);
      //MPI_Send_mine (buff2,buffsize2,MPI_DOUBLE,0,Myrank,MPI_COMM_WORLD);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    fflush (stdout);
    // **************************************************************************
    
    if (Myrank==0){
      // *****************************************
      //  solution of TRANSPORT macroscale problem
      // *****************************************
      
      if((Tp->tprob == nonstationary_problem) || (Tp->tprob == growing_np_problem)){
	// linear algorithm
	// perform one time step with linear solver
	tret = one_step_linear(tlcid, tnewtime, tdt, prev_timet, rest_calct, ti, tli, np_gv);
	trfel_convergence = 1;
      }
      
      if(Tp->tprob == discont_nonstat_problem){
	print_err("the solver has not yet been implemented", __FILE__, __LINE__, __func__);
	// perform one time step with linear solver
	// tret = one_step_linear(tlcid, tnewtime, tdt, ti, tli, np_gv);
	// trfel_convergence = 1;
      }
      
      if(Tp->tprob == discont_nonlin_nonstat_problem){
	// perform one time step with nonlinear solver
	tret = one_step_nonlinear_dform(tlcid, tnewtime, tdt, ti, tli, np_gv);
	trfel_convergence = 1;
      }
      
      if((Tp->tprob == nonlinear_nonstationary_problem) || (Tp->tprob == growing_np_problem_nonlin)){
	// non-linear algorithm
	// perform one time step with non-linear solver
	tret = one_step_nonlinear(tlcid, tnewtime, tdt, prev_timet, rest_calct, ti, tli, np_gv);
	trfel_convergence = 1;
      }
      
      //  handling with time controler for nonlinear solver
      if ((((Tp->tprob == nonlinear_nonstationary_problem) || (Tp->tprob == growing_np_problem_nonlin)) || 
	   (Tp->tprob == discont_nonlin_nonstat_problem))  && Tp->timecont.tct > 0)
	{
	  if (tret >= 0){
	    // equilibrium has been attained
	    
	    //  approximation of nodal values into integration points
	    approximation ();
	    if (Cp->dpt == pass_by_aux_ip)
	      aip_approximation(Mm->tnip, MTipmap);
	    Tm->updateipval();
	    //Tm->freezing_thawing ();
	    
	    trfel_convergence = 1;
	    
	    if ((tret == 0) || (tret < Tp->nii/4))
	      tnsts++;
	    else
	      tnsts = btnsts = 0; 
	    
	    if (tnsts>1){
	      if (tdt < Tp->timecont.dtmax){
		tdt*=2.0;
		if (tdt > Tp->timecont.dtmax)
		  tdt = Tp->timecont.dtmax;
		tnsts=0;
		btnsts=1;
		
		if (Mesprt==1)  
		  fprintf (stdout,"\n\n TRFEL time increment may be enlarged because no inner loop was neccessary in previous steps");
	      }
	    }
	  }
	  else{
	    // equilibrium has not been attained
	    tnsts = btnsts = 0;
	    
	    //  reduction of the time increment because
	    //  inner loop was not able to enforce equilibrium
	    trfel_convergence = 0;
	    
	    tdt/=2.0;
	    ti--;
	    
	    if(Tp->timecont.tct > 0) // for adaptive time controler
	      {
		//restoring results from previous time step
		copyv(np_gv.lhsb, np_gv.lhs, Ndoft);
		copyv(np_gv.tdlhsb, np_gv.tdlhs, Ndoft);
	      }
	    //  approximation of nodal values into integration points
	    approximation ();
	    if (Cp->dpt == pass_by_aux_ip)
	      aip_approximation(Mm->tnip, MTipmap);
	    
	    if (Mesprt==1)  
	      fprintf (stdout,"\n\n TRFEL time increment is reduced to dt=%le because the inner loop was not able to enforce equilibrium", tdt);
	    
	    if (tdt<Tp->timecont.dtmin){
	      if (Mesprt==1)  fprintf (stderr," TRFEL time increment is less than the minimum time increment");
	      if (Mesprt==1)  fprintf (stderr,"\n computation failed (file %s, line %d)\n",__FILE__,__LINE__);
	      if (Mesprt==1)  fprintf (stderr,"\n FORCED output of results from this step is performed\n");
	      
	      compute_req_valt (lcid);
	      print_stept_forced(lcid, ti, Tp->time, np_gv.rhs);
	      print_flusht();
	      
	      compute_req_val (mlcid);
	      print_step_forced(mlcid, mi, Mp->time, mt_gv.fb);
	      print_flush();
	      
	      stop=1;
	      break;
	    }
	    
	    Tp->timecont.oldtime ();
	    Mp->timecon.oldtime ();
	    
	    // change status of elements and nodes according to the previous time step
	    if ((Tp->tprob == growing_np_problem) || (Tp->tprob == growing_np_problem_nonlin))
	      {
		update_elemnod_prev_timet(prev_timet, np_gv.ncd, np_gv.nce);
	      }
	    
	    continue;
	  }
	}
      
      
      if (stop==1){
	//  TRFEL requires smaller time step than the minimum time step
	//  the code will be terminated
	break;
      }
      
      //  time synchronization
      if (Mp->time > Tp->time){
	Mp->timecon.oldtime ();
	mdt=Tp->time - Mp->timecon.time;
	mnewtime = Mp->timecon.newtime (mdt);
	Mp->time = mnewtime;
	if (bmnsts == 1)
	  mnsts = bmnsts;
      }
      
      //  METR cleans TRFEL matrices for more memory in MEFEL part
      if (Cp->cleanmatrix == clean_yes){
	//  cleaning matrices for more free memmory
	//  cleaning of the conductivity matrix
	if (Kmat != NULL){
	  fprintf (stdout,"\n\n Cleaning of conductivity matrix \n\n ");
	  delete Kmat;
	  Kmat=NULL;
	}
	//  cleaning of the capacity matrix
	if (Cmat != NULL){
	  fprintf (stdout,"\n\n Cleaning of capacity matrix \n\n ");
	  delete Cmat;
	  Cmat=NULL;
	}
      }
      
      //  MEFEL solution
      mefel_convergence = 0;
      //  new mefel step number
      mi++;
      
      //  copying of data
      //  copying of transport data to MEFEL
      trfel_mefel ();
    }
    // **************************************************************************

    if (Myrank==0){
      // *******************************************
      //  solution of microscale MECHANICAL problems
      // *******************************************
      for (ii=1;ii<Nproc;ii++){
	
	//  function assembles data from the macroproblem for the ii-th microproblem to the array buff1
	// tady pole buff1 musi mit rozmer podle nelidomm??!! nebo podle poctu elementu??!!
	higher_to_lower_levelm (ii,buffsize3,buff3);
	
	MPI_Send (buff3,buffsize3,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
	//MPI_Send_mine (buff3,buffsize3,MPI_DOUBLE,ii,Myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Recv (buff3,buffsize3,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      //MPI_Recv_mine(buff3,buffsize3,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      for (ii=0;ii<nelidomm;ii++){
	
	/* fprintf (Out,"\n\n *********************************************");
	   fprintf (Out,"\n Myrank  = %ld",Myrank);
	   fprintf (Out,"\n nelidomm = %ld",ii);
	   fprintf (Out,"\n\n pred vypoctem paral_mechanical_homogenization:");
	   for (long ijk=0;ijk<buffsize3;ijk++){
	   //for (long ijk=0;ijk<1;ijk++){
	   fprintf (Out,"\n buff3 %ld   %le",ijk,buff3[ijk]);
	   }
	   
	   fprintf (Out,"\n");
	   for (long ijk=0;ijk<buffsize4;ijk++){
	   //for (long ijk=0;ijk<1;ijk++){
	   fprintf (Out,"\n buff4 %ld   %le",ijk,buff4[ijk]);
	   }
	   fflush(Out);
	*/

	paral_mechanical_homogenization (buff3+ii*(dimm+ncomp+1),buff4+ii*(ncomp*ncomp));
	
	/* fprintf (Out,"\n\n po vypoctu paral_mechanical_homogenization:");
	   for (long ijk=0;ijk<buffsize3;ijk++){
	   //for (long ijk=0;ijk<1;ijk++){
	   fprintf (Out,"\n buff3 %ld   %le",ijk,buff3[ijk]);
	   }
	   fprintf (Out,"\n");
	   for (long ijk=0;ijk<buffsize4;ijk++){
	   //for (long ijk=0;ijk<1;ijk++){
	   fprintf (Out,"\n buff4 %ld   %le",ijk,buff4[ijk]);
	   }
	   fflush(Out);
	*/
      }
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    // **************************************************************************
    
    if (Myrank==0){
      for (ii=1;ii<Nproc;ii++){
	MPI_Recv (buff4,buffsize4,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	//MPI_Recv_mine(buff4,buffsize4,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	
      	k=stat.MPI_TAG;
	jj=0;
	
	for (j=0;j<Mt->ne;j++){
	  
	  if (Gtm->stop->eldom[j]==k){
	    //  only elements connected with the k-th microscale are assumed
	  
	    /* fprintf (Out,"\n");
	       fprintf (Out,"\n Myrank  = %ld",Myrank);
	       fprintf (Out,"\n element no. = %ld",j);
	       fprintf (Out,"\n Gtm->stop->eldom[j] = %ld",Gtm->stop->eldom[j]);
	       fprintf (Out,"\n ii = %ld",ii);
	       fprintf (Out,"\n k  =stat.MPI_TAG = %ld",k);
	       fprintf (Out,"\n");
	       fflush(Out);
	    */

	    switch (Mp->mami){
	    case elem_domain:{
	      //  each element is connected with its own microproblem

	      Mm->hommatm[j].assemble_matrices (buff4,ncomp,dimm);
	      break;
	    }
	      
	    case aggregate_domain:{
	      //  generation of auxiliary data needed in parallel computation
	      Mm->hommatm[j].assemble_matrices (buff4+jj*(ncomp*ncomp),ncomp,dimm);
	      jj++;
	      break;
	    }
	    case material_aggregate_domain:{
	      //  aggregate sends averaged data to microproblem
	      //  index jj is not incremented in this case because all elements in aggregate??!!
	      //  obtain the same conductivity and capacity matrices
	      //
	      //  index of subdomain-aggregate = index of material (processor)
	      
	      //fprintf (Outt,"\n Myrank  = %ld",Myrank);
	      //fprintf (Outt,"\n from Myrank  = %ld",ii);
	      //fprintf (Outt,"\n element  = %ld",j);
	      //fprintf (Outt,"\n\n pred ulozenim matic na materialu c. %ld:",j);
	      //for (long ijk=0;ijk<buffsize2;ijk++){
	      //fprintf (Outt,"\n buff4 %ld   %le",ijk,buff4[ijk]);
	      //}
	      
	      Mm->hommatm[j].assemble_matrices (buff4,ncomp,dimm);

	      //fflush(Outt);
	      break;
	    }
	    case eff_aggregate_domain:{
	      //  aggregate sends averaged data to microproblem
	      //  index jj is not incremented in this case because all elements in aggregate
	      //  obtain the same conductivity and capacity matrices
	      //
	      //  index of subdomain-aggregate = index of element material
	      kk = Mt->elements[i].idm[0];
	      Mm->hommatm[kk].assemble_matrices (buff4+kk*(ncomp*ncomp),ncomp,dimm);
	      break;
	    }
	      
	    default:{
	      par_print_err(Myrank,"unknown type of aggregation is required",__FILE__,__LINE__,__func__);
	    }
	    } 
	  }
	}//  end loop over the number of elements
      }//  end loop over the number of processors
    }//  end of the master processor part
    else{
      MPI_Send (buff4,buffsize4,MPI_DOUBLE,0,Myrank,MPI_COMM_WORLD);
      //MPI_Send_mine (buff4,buffsize4,MPI_DOUBLE,0,Myrank,MPI_COMM_WORLD);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    fflush (stdout);
   

    // **************************************************************************
   
    if (Myrank==0){
      
      // ******************************************
      //  solution of MECHANICAL macroscale problem
      // ******************************************
      
      if ((Mp->tprob == growing_mech_structure) && (prev_timem == Cp->timecon.starttime()))
	prev_timem = Mp->time;
      //  perform one time step
      //mret = parallel_homogenization_one_step (mlcid, mnewtime, mdt, mdtr, prev_timem, rest_calcm, mi, mli, mt_gv);
      mret = one_step (mlcid, mnewtime, mdt, mdtr, prev_timem, rest_calcm, mi, mli, mt_gv);
      
      if (mret >= 0){
	//  equilibrium has been attained
	
	if ((mret == 0) || (mret < Mp->nlman->niilnr/4))
	  mnsts++;      
	else
	  mnsts = bmnsts = 0;
	
	if (mnsts > 1){
	  if (mdt < Mp->timecon.dtmax){
	    mdt*=2.0;
	    if (mdt > Mp->timecon.dtmax)
	      mdt = Mp->timecon.dtmax;
	    mnsts=0;
	    if (Mespr==1)  
	      fprintf (stdout,"\n\n MEFEL time increment is enlarged because no inner loop was neccessary in previous steps\n");
	  }
	}
      }
      else{
	// equilibrium has not been attained
	mnsts = bmnsts = 0;
	
	//  reduction of the time increment because
	//  inner loop was not able to enforce equilibrium
	mefel_convergence = 0;
	
	//  reduction of the time increment because
	//  inner loop was not able to enforce equilibrium
	if (mdtr < 1.0)
	  mdt *= mdtr;
	else
	  mdt/=2.0;
	
	Mp->timecon.oldtime ();
	Tp->timecont.oldtime ();
	mi--;
	ti--;	
	if(Tp->timecont.tct > 0) // for adaptive time controler
	  {
	    //restoring results from previous time step
	    copyv(np_gv.lhsb, np_gv.lhs, Ndoft);
	    copyv(np_gv.tdlhsb, np_gv.tdlhs, Ndoft);
	  }
	//  approximation of nodal values into integration points
	approximation ();
	if (Cp->dpt == pass_by_aux_ip)
	  aip_approximation(Mm->tnip, MTipmap);
	
	// copy state variables from the last attained equilibrium state (eqother array) to the actual state variables (other array)
	Mm->updateother();
	
	if (Mespr==1)  
	  fprintf (stdout,"\n\n MEFEL time increment is reduced because the inner loop was not able to enforce equilibrium");
	if (mdt<Mp->timecon.dtmin){
	  if (Mespr==1)  fprintf (stderr, " MEFEL time increment is less than the minimum time increment");
	  if (Mespr==1)  fprintf (stderr, "\n computation failed (file %s, line %d)\n",__FILE__,__LINE__);
	  if (Mespr==1)  fprintf (stderr, "\n FORCED output of results from this step is performed\n");
	  
	  
	  compute_req_valt (lcid);
	  print_stept_forced(lcid, ti, Tp->time, np_gv.rhs);
	  print_flusht();
	  
	  compute_req_val (mlcid);
	  print_step_forced(mlcid, mi, Mp->time, mt_gv.fb);
	  print_flush();
	  
	  stop=1;
	  break;
	}
	// change status of elements and nodes according to the previous time step
	if (Mp->tprob == growing_mech_structure)
	  {
	    update_elemnod_prev_time(prev_timem, mt_gv.ncd, mt_gv.nce);
	  }
	continue;
      }
      
      if (stop==1){
	//  MEFEL requires smaller time step than the minimum time step
	//  the code will be terminated
	break;
      }
      
      //    
      // Equilibrium was attained both in TRFEL and MEFEL
      //
      
      // For nonlinear TRFEL problems
      //
      // backup nodal values, lshsb = lhs
      copyv(np_gv.lhs, np_gv.lhsb, Ndoft);
      // backup time derivatives of nodal values, tdlhsb = tdlhs
      copyv(np_gv.tdlhs, np_gv.tdlhsb, Ndoft);
      
      //  printing of TRFEL output and graphical informations
      compute_req_valt (lcid);
      print_stept(lcid,Tp->istep,Tp->time,np_gv.rhs);//print_stept(0,i,Tp->time,NULL);
      print_flusht();
      // perform TRFEL backup
      if ((Tp->timecont.isitimptime ()==1) && Tp->hdbcont.save_stat())
	{
	  if (Mesprt==1)
	    fprintf (stdout,"\n Creating TRFEL backup file\n");
	  solvert_save (np_gv.lhs,np_gv.tdlhs,np_gv.f,Tp->istep,Tp->time,tdt,Tp->timecont,Ndoft);
	}
      
      // for TRFEL nonlinear problems, save vectors of unknowns and unknown increments
      if (((Tp->tprob == nonlinear_nonstationary_problem) || (Tp->tprob == growing_np_problem_nonlin) || 
	   (Tp->tprob == discont_nonlin_nonstat_problem))  && Tp->timecont.tct > 0)
	{
	  // storage of results from previous inner step
	  copyv(np_gv.lhs, np_gv.lhsb, Ndoft);
	  copyv(np_gv.tdlhs, np_gv.tdlhsb, Ndoft);
	}
      
      if ((Tp->tprob == growing_np_problem) || (Tp->tprob == growing_np_problem_nonlin))
	{
	  // Following two commands must be performed here because in case that
	  // elements or nodes are changed in the next step and the step would not converge,
	  // the lhs and tdlhs could not be recovered correctly
	  
	  // backup of nodal values because code numbers may be changed
	  //debug??!!Tt->lhs_save (np_gv.lhs,Lsrst->lhsi,np_gv.tdlhs);
	  //  backup of nodal forces because code numbers may be changed//from mefel??!!
	  //Mt->save_nodforce(mt_gv.f);//from mefel???!!!
	  
	  // new elements can be definitely changed to old ones
	  Gtt->update_auxinf();
	  prev_timet = Tp->time;
	}
      
      // for MEFEL nonlinear problems
      //  Backup of r is used in case that the Newton-Raphson procedure does not converge.
      //  lhsb[j]=r[j]
      copyv(mt_gv.r, mt_gv.lhsb, Ndofm);
      // actual total load vector becomes previous total load vector
      // fp[k]=f[k]
      copyv(mt_gv.f, mt_gv.fp, Ndofm);
      
      if (Mp->tprob == growing_mech_structure)
	{
	  // Following two commands must be performed here because in case that
	  // elements or nodes are changed in the next step and the step would not converge,
	  // the nodal values and nodal forces could not be recovered correctly
	  // nodval due to changes in Mt->nodedispl and nodforce due to changes in vector f
	  
	  // backup of nodal values because code numbers may be changed due to
	  Mt->save_nodval(lcid);
	  //  backup of nodal forces because code numbers may be changed
	  Mt->save_nodforce(mt_gv.f);
	  
	  // new elements can be definitely changed to old ones
	  Gtm->update_auxinf();
	  prev_timem = Mp->time;
	}
      
      //  update of values stored at integration points
      Mm->updateipval();
      
      mefel_trfel(lcid);
      
      //  printing of MEFEL output and graphical informations
      compute_req_val (lcid);
      print_step(lcid, Mp->istep, Mp->time, mt_gv.fl);
      print_flush();
      
      // perform MEFEL backup
      if ((Mp->timecon.isitimptime()==1) && (Mp->hdbcont.save_stat()))
	{
	  if (Mespr==1)
	    fprintf (stdout,"\n Creating backup file\n");
	  solver_save (mt_gv.r,mt_gv.fp,Mp->istep,Mp->time,dt,&Mp->timecon,Ndofm);
	}
      
      //  METR cleans MEFEL matrices for more memory in TRFEL part
      if(Cp->cleanmatrix == clean_yes){
	//  cleaning of stiffness matrix
	if (Smat != NULL){
	  fprintf (stdout,"\n\n Cleaning of stiffness matrix \n\n ");
	  delete Smat;
	  Smat=NULL;
	}
      }
      
      if ((Tp->tprob == growing_np_problem) || (Tp->tprob == growing_np_problem_nonlin))
	prev_timet = tnewtime;
    }

    fprintf (stdout,"\n procesor %d time = %lf end_time = %lf",Myrank,time,end_time);
    fflush (stdout);
    
    // **************************************************************************

  }while (Cp->time < end_time); 
 
  if (Myrank==0){
    print_close ();
    print_closet();
  }

  delete [] buff1;
  delete [] buff2; 
  delete [] buff3;
  delete [] buff4;

}

